"""Implementation of the seam carving algorithm."""
import numpy as np


def horizontal_shrink(
        image: np.ndarray, derivative: np.ndarray,
) -> np.ndarray:
    """Horizontal image compression.

    :param image: Input image.
    :type image: np.ndarray
    :param derivative: The module of the image derivative.
    :type derivative: np.ndarray
    :return: Output image.
    :rtype: np.ndarray
    """
    return shrink(image, derivative)


def vertical_shrink(
        image: np.ndarray, derivative: np.ndarray,
) -> np.ndarray:
    """Vertical image compression.

    :param image: Input image.
    :type image: np.ndarray
    :param derivative: The module of the image derivative.
    :type derivative: np.ndarray
    :return: Output image.
    :rtype: np.ndarray
    """
    image = shrink(
        np.transpose(image, axes=(1, 0, 2)), np.transpose(derivative),
    )
    return np.transpose(image, axes=(1, 0, 2))


def compute_seam(derivative: np.ndarray) -> np.ndarray:
    """Calculate the seam with the least energy.

    :param derivative: The module of the image derivative.
    :type derivative: np.ndarray
    :return: A boolean image mask with a seam whose energy is minimal.
    :rtype: np.ndarray
    """
    energy = np.zeros(derivative.shape)
    energy[0] = derivative[0]

    for height in range(1, energy.shape[0]):
        for width in range(energy.shape[1]):
            if width not in (0, energy.shape[1] - 1):
                energy[height, width] = (
                    min(
                        energy[height - 1, width - 1],
                        energy[height - 1, width],
                        energy[height - 1, width + 1],
                    ) + derivative[height, width]
                )
            elif width != 0 and width == energy.shape[1] - 1:
                energy[height, width] = (
                    min(
                        energy[height - 1, width - 1],
                        energy[height - 1, width],
                    ) + derivative[height, width]
                )
            elif width == 0 and width != energy.shape[1] - 1:
                energy[height, width] = (
                    min(
                        energy[height - 1, width],
                        energy[height - 1, width + 1],
                    ) + derivative[height, width]
                )
            else:
                energy[height, width] = (
                    energy[height - 1, width] + derivative[height, width]
                )

    seam = np.zeros(energy.shape)
    index = np.argmin(energy[-1])

    seam[-1, index] = True

    for height in range(energy.shape[0] - 2, -1, -1):
        if index not in (0, energy.shape[1] - 1):
            index += np.argmin(energy[height, index - 1: index + 2]) - 1
        elif index != 0 and index == energy.shape[1] - 1:
            index += np.argmin(energy[height, index - 1: index + 1]) - 1
        elif index == 0 and index != energy.shape[1] - 1:
            index += np.argmin(energy[height, index: index + 2])

        seam[height, index] = True

    return seam


def shrink(
        image: np.ndarray, derivative: np.ndarray,
) -> np.ndarray:
    """Image compression.

    :param image: Input image.
    :type image: np.ndarray
    :param derivative: The module of the image derivative.
    :type derivative: np.ndarray
    :return: Output image.
    :rtype: np.ndarray
    """
    seam = compute_seam(derivative)

    red = np.ma.compressed(np.ma.MaskedArray(image[:, :, 0], mask=seam))
    green = np.ma.compressed(np.ma.MaskedArray(image[:, :, 1], mask=seam))
    blue = np.ma.compressed(np.ma.MaskedArray(image[:, :, 2], mask=seam))

    red = red.reshape(image.shape[0], image.shape[1] - 1)
    green = green.reshape(image.shape[0], image.shape[1] - 1)
    blue = blue.reshape(image.shape[0], image.shape[1] - 1)

    image = np.dstack((red, green, blue))

    return image


FUNCTIONS = {
    'horizontal': horizontal_shrink,
    'vertical': vertical_shrink,
}


def seam_carve(
        image: np.ndarray, action: str,
) -> np.ndarray:
    """Apply a seam carving algorithm.

    :param image: Input image.
    :type image: np.ndarray
    :param action: The direction of image compression horizontal or vertical.
    :type action: str
    :return: Output image.
    :rtype: np.ndarray
    """
    brightness = (
        image[:, :, 0] * 0.299 + image[:, :, 1] * 0.587 + image[:, :, 2] * 0.114
    )

    append_up = np.append(brightness[0][np.newaxis], brightness, axis=0)[:-1]
    append_down = np.append(brightness, brightness[-1][np.newaxis], axis=0)[1:]

    append_left = np.append(
        brightness[:, 0][:, np.newaxis], brightness, axis=1,
    )[:, :-1]
    append_right = np.append(
        brightness, brightness[:, -1][:, np.newaxis], axis=1,
    )[:, 1:]

    derivative_x = append_right - append_left
    derivative_y = append_down - append_up
    derivative = np.sqrt(derivative_x ** 2 + derivative_y ** 2)

    return FUNCTIONS[action](image, derivative)



import os
import unittest

from jackalify.app import app
from fastapi.testclient import TestClient


class TestApi(unittest.TestCase):

    def setUp(self) -> None:
        self.data_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))
        self.client = TestClient(app)
        self.result_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                       "..", "jackalify", "static", "working"))

    def test_correct_file(self):
        response = self.client.get("/")
        self.assertEqual(200, response.status_code)
        if os.path.isfile(os.path.join(self.result_dir, 'jackalified.gif')):
            os.remove(os.path.join(self.result_dir, 'jackalified.gif'))
        if os.path.isfile(os.path.join(self.result_dir, 'jackalified.png')):
            os.remove(os.path.join(self.result_dir, 'jackalified.png'))

        with open(os.path.join(self.data_dir, f"black.png"), "rb") as f:
            files = {"file": (f.name, f, "multipart/form-data")}
            response = self.client.post("/", files=files)
            self.assertEqual(200, response.status_code)
            self.assertTrue(os.path.isfile(os.path.join(self.result_dir, 'jackalified.gif')))
            self.assertTrue(os.path.isfile(os.path.join(self.result_dir, 'jackalified.png')))

    def test_wrong_file(self):
        response = self.client.get("/")
        self.assertEqual(200, response.status_code)
        with open(os.path.join(self.data_dir, f"coloured_jackalified.gif"), "rb") as f:
            files = {"file": (f.name, f, "multipart/form-data")}
            response = self.client.post("/", files=files)
            self.assertEqual(415, response.status_code)



import os
import unittest
from copy import deepcopy

import cv2
import numpy as np

from jackalify.jackal import jackalify
from jackalify.seam_carve import horizontal_shrink, vertical_shrink, shrink, seam_carve


class TestBackend(unittest.TestCase):

    def setUp(self) -> None:
        self.data_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))
        self.img_name = "black"

    def test_seam_carve(self):
        test_array = cv2.imread(os.path.join(self.data_dir, f"{self.img_name}.png"))
        derivative = deepcopy(test_array[:, :, 0])
        self.assertNotEqual(id(derivative), id(test_array))

        res_array = shrink(test_array, derivative)
        self.assertEqual((test_array.shape[0], test_array.shape[1] - 1, test_array.shape[2]), res_array.shape)
        res_array = vertical_shrink(test_array, derivative)
        self.assertEqual((test_array.shape[0] - 1, test_array.shape[1], test_array.shape[2]), res_array.shape)
        res_array = horizontal_shrink(test_array, derivative)
        self.assertEqual((test_array.shape[0], test_array.shape[1] - 1, test_array.shape[2]), res_array.shape)
        res_array = seam_carve(test_array, 'horizontal')
        self.assertEqual((test_array.shape[0], test_array.shape[1] - 1, test_array.shape[2]), res_array.shape)
        res_array = seam_carve(test_array, 'vertical')
        self.assertEqual((test_array.shape[0] - 1, test_array.shape[1], test_array.shape[2]), res_array.shape)

    def test_jackalify(self):
        jackalify(os.path.join(self.data_dir, f"{self.img_name}.png"),
                  os.path.join(self.data_dir, f"{self.img_name}.gif"))
        self.assertTrue(os.path.isfile(os.path.join(self.data_dir, f"{self.img_name}.gif")))
        os.remove(os.path.join(self.data_dir, f"{self.img_name}.gif"))

        jackalify(os.path.join(self.data_dir, f"{self.img_name}.png"),
                  os.path.join(self.data_dir, f"{self.img_name}.gif"),
                  os.path.join(self.data_dir, f"{self.img_name}_out.png"))
        self.assertTrue(os.path.isfile(os.path.join(self.data_dir, f"{self.img_name}.gif")))
        self.assertTrue(os.path.isfile(os.path.join(self.data_dir, f"{self.img_name}_out.png")))
        os.remove(os.path.join(self.data_dir, f"{self.img_name}.gif"))
        os.remove(os.path.join(self.data_dir, f"{self.img_name}_out.png"))

    def test_result(self):
        jackalify(os.path.join(self.data_dir, "coloured.png"),
                  os.path.join(self.data_dir, "coloured.gif"),
                  os.path.join(self.data_dir, "coloured_result.png"))
        test_img = cv2.imread(os.path.join(self.data_dir, "coloured_jackalified.png"))
        result_img = cv2.imread(os.path.join(self.data_dir, "coloured_result.png"))
        self.assertTrue(np.allclose(test_img, result_img))
        os.remove(os.path.join(self.data_dir, "coloured.gif"))
        os.remove(os.path.join(self.data_dir, "coloured_result.png"))
